<?php
/**
 * The admin-specific functionality of the plugin.
 *
 * @link       https://club.wpeka.com
 * @since      7.0
 *
 * @package    Wplegalpages_Pro
 * @subpackage Wplegalpages_Pro/admin
 */

/**
 * The admin-specific functionality of the plugin.
 *
 * Defines the plugin name, version, and two examples hooks for how to
 * enqueue the admin-specific stylesheet and JavaScript.
 *
 * @package    Wplegalpages_Pro
 * @subpackage Wplegalpages_Pro/admin
 * @author     WPEkaClub <support@wpeka.com>
 */
class Wplegalpages_Pro_Admin {

	/**
	 * The ID of this plugin.
	 *
	 * @since    7.0
	 * @access   private
	 * @var      string    $plugin_name    The ID of this plugin.
	 */
	private $plugin_name;

	/**
	 * The version of this plugin.
	 *
	 * @since    7.0
	 * @access   private
	 * @var      string    $version    The current version of this plugin.
	 */
	private $version;

	/**
	 * Initialize the class and set its properties.
	 *
	 * @since    7.0
	 * @param      string $plugin_name       The name of this plugin.
	 * @param      string $version    The version of this plugin.
	 */
	public function __construct( $plugin_name, $version ) {

		$this->plugin_name = $plugin_name;
		$this->version     = $version;

		$lp_general = get_option( 'lp_general' );
		if ( isset( $lp_general['affiliate-disclosure'] ) && '1' === $lp_general['affiliate-disclosure'] ) {
			add_action( 'add_meta_boxes', array( $this, 'wplegalpages_pro_register_meta_boxes' ) );
			add_action( 'save_post', array( $this, 'wplegalpages_pro_save_meta_box' ) );
			add_filter( 'the_content', array( $this, 'wplegalpages_pro_post_content' ) );
		}

	}

	/**
	 * Register the stylesheets for the admin area.
	 *
	 * @since    7.0
	 */
	public function enqueue_styles() {

		/**
		 * This function is provided for demonstration purposes only.
		 *
		 * An instance of this class should be passed to the run() function
		 * defined in Wplegalpages_Pro_Loader as all of the hooks are defined
		 * in that particular class.
		 *
		 * The Wplegalpages_Pro_Loader will then create the relationship
		 * between the defined hooks and the functions defined in this
		 * class.
		 */

		wp_enqueue_style( $this->plugin_name, plugin_dir_url( __FILE__ ) . 'css/wplegalpages-pro-admin' . WPLPP_SUFFIX . '.css', array(), $this->version, 'all' );

	}

	/**
	 * Register the JavaScript for the admin area.
	 *
	 * @since    7.0
	 */
	public function enqueue_scripts() {

		/**
		 * This function is provided for demonstration purposes only.
		 *
		 * An instance of this class should be passed to the run() function
		 * defined in Wplegalpages_Pro_Loader as all of the hooks are defined
		 * in that particular class.
		 *
		 * The Wplegalpages_Pro_Loader will then create the relationship
		 * between the defined hooks and the functions defined in this
		 * class.
		 */

		wp_enqueue_script( $this->plugin_name, plugin_dir_url( __FILE__ ) . 'js/wplegalpages-pro-admin' . WPLPP_SUFFIX . '.js', array( 'jquery' ), $this->version, false );

	}

	/**
	 * Registeres Affiliate disclosure metabox for posts.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_register_meta_boxes() {
		add_meta_box( 'meta-box-id', esc_attr__( 'Add Affiliate Disclosure', 'wplegalpages' ), array( $this, 'wplegalpages_pro_display_callback' ), 'post' );
	}

	/**
	 * Save post meta data.
	 *
	 * @param int $post_id Post ID.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_save_meta_box( $post_id ) {
		if ( isset( $_POST['wplegal_custom_metadata'] ) ) {
			check_admin_referer( 'wplegal_save_custom_metabox', 'wplegal_custom_metadata' );
		}
		$legal_content = isset( $_POST['legal_content'] ) ? wp_kses_post( wp_unslash( $_POST['legal_content'] ) ) : '';
		$legal_check   = isset( $_POST['legal_check'] ) ? sanitize_text_field( wp_unslash( $_POST['legal_check'] ) ) : '';
		update_post_meta( $post_id, '_legal_content', $legal_content );
		update_post_meta( $post_id, '_legal_check', $legal_check );
	}

	/**
	 * Modify post content to include affiliate disclosure.
	 *
	 * @param string $content Post content.
	 *
	 * @return string
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_post_content( $content ) {
		global $post;
		$legal_check_status = get_post_meta( $post->ID, '_legal_check', true );
		$meta               = get_post_meta( $post->ID, '_legal_content', true );
		$meta               = "<div id='_affiliate_disclosure'><i>" . $meta . '</i></div>';
		if ( 'on' === $legal_check_status ) {
			$content = $content . $meta;
		}
		return $content;
	}

	/**
	 * Affiliate disclosure content for post.
	 *
	 * @param WP_Post $post Post.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_display_callback( $post ) {
		$legal_check_status = get_post_meta( $post->ID, '_legal_check', true );
		$legal_content      = get_post_meta( $post->ID, '_legal_content', true );
		wp_nonce_field( 'wplegal_save_custom_metabox', 'wplegal_custom_metadata' );
		if ( '' === $legal_content ) {
			$legal_content = "Disclosure: Some of the links in this post are 'affiliate links.' This means if you click on the link and purchase the item, I will receive an affiliate commission.";
		}

		?>

		<script>
			jQuery(document).ready(function(){

				jQuery("#legal_check").click(function(){

					if(jQuery("#legal_check").is(':checked'))
					{
						jQuery("#content_area_read").hide();
						jQuery("#content_area_edit").show();
					}
					else
					{
						jQuery("#content_area_read").show();
						jQuery("#content_area_edit").hide();
					}

				});
			});
		</script>

		<table>
			<tr>
				<?php if ( 'on' !== $legal_check_status ) { ?>
					<td valign="middle" width="10%" align="center"> <input type="checkbox" name="legal_check" id="legal_check" > </td>
					<td width="90%">
						<div id="content_area_edit" style="display:none;">
							<textarea name="legal_content" cols="80" rows="2" ><?php echo esc_attr( $legal_content ); ?></textarea>
						</div>
						<div id="content_area_read">
							<p><?php echo esc_attr( $legal_content ); ?></p>
						</div>
					</td>
				<?php } else { ?>
					<td valign="middle" width="10%" align="center"> <input type="checkbox" checked name="legal_check" id="legal_check" > </td>
					<td width="90%">
						<div id="content_area_edit">
							<textarea name="legal_content" cols="80" rows="2" ><?php echo esc_attr( $legal_content ); ?></textarea>
						</div>
						<div id="content_area_read" style="display:none;">
							<p><?php echo esc_attr( $legal_content ); ?></p>
						</div>
					</td>
				<?php } ?>

			</tr>
		</table>

		<?php
	}

	/**
	 * Legal Pages widget.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_widget_dashboard() {
		echo '<span style="line-height:18px;">WP Legal Pages plugin is the easiest method to create legal pages for your website.</span>';
	}

	/**
	 * Legal Pages widget setup.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_dashboard_setup() {
		wp_add_dashboard_widget( 'lp_dashboard_widget', 'Legal Pages', array( $this, 'wplegalpages_pro_widget_dashboard' ) );
	}

	/**
	 * Admin menu for Legal Pages Pro.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_admin_menu() {
		$terms = get_option( 'lp_accept_terms' );
		if ( '1' === $terms ) {
			add_submenu_page(
				'legal-pages',
				'Create/Edit Templates',
				esc_attr__( 'Create/Edit Templates', 'wplegalpages' ),
				'manage_options',
				'lp-templates',
				array(
					$this,
					'create_templates',
				)
			);
			add_submenu_page(
				'legal-pages',
				'Create Popups',
				esc_attr__( 'Create Popups', 'wplegalpages' ),
				'manage_options',
				'lp-popups',
				array(
					$this,
					'create_popups',
				)
			);
			add_submenu_page(
				'legal-pages',
				'Age verification',
				esc_attr__( 'Age verification', 'wplegalpages' ),
				'manage_options',
				'lp-age-verify',
				array(
					$this,
					'age_verify',
				)
			);
		}
	}

	/**
	 * Create/Edit templates.
	 *
	 * @since 7.0
	 */
	public function create_templates() {
		include_once plugin_dir_path( dirname( __FILE__ ) ) . 'admin/create-templates.php';
	}

	/**
	 * Create popups.
	 *
	 * @since 7.0
	 */
	public function create_popups() {
		include_once plugin_dir_path( dirname( __FILE__ ) ) . 'admin/create-popups.php';
	}

	/**
	 * Age verification admin form.
	 *
	 * @since 7.0
	 */
	public function age_verify() {
		include_once plugin_dir_path( dirname( __FILE__ ) ) . 'admin/age-verify.php';
	}

	/**
	 * Legal Pages Pro admin settings.
	 *
	 * @param Array $lp_general Setting options.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_admin_settings( $lp_general ) {
		include_once plugin_dir_path( dirname( __FILE__ ) ) . 'admin/admin-settings.php';
	}

	/**
	 * Limit creation of LegalPages.
	 *
	 * @param int $max_limit Maximum limit to create legal pages.
	 *
	 * @return int
	 *
	 * @since 7.1
	 */
	public function wplegalpages_pro_pages_limit( $max_limit ) {
		$max_limit = 100;
		return $max_limit;
	}

	/**
	 * Get pages excluded.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_init() {
		add_filter( 'get_pages', array( $this, 'wplegalpages_pro_exclude_pages' ) );
		$locale  = get_locale();
		$folder  = rtrim( basename( dirname( __FILE__ ) ), '/' );
		$mo_file = trailingslashit( WP_PLUGIN_DIR ) . "$folder/languages/wplegalpages-$locale.mo";
		load_textdomain( 'wplegalpages', $mo_file );
	}

	/**
	 * Get excluded pages.
	 *
	 * @param Array $pages Pages.
	 *
	 * @return array
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_exclude_pages( $pages ) {
		$baseurl  = isset( $_SERVER['PHP_SELF'] ) ? sanitize_text_field( wp_unslash( $_SERVER['PHP_SELF'] ) ) : '';
		$bail_out = ( ( defined( 'WP_ADMIN' ) && true === WP_ADMIN ) || ( false !== strpos( $baseurl, 'wp-admin' ) ) );
		$bail_out = apply_filters( 'wplegalpages_pro_admin_bail_out', $bail_out );
		if ( $bail_out ) {
			return $pages;
		}
		$excluded_ids = $this->wplegalpages_pro_get_excluded_ids();
		$length       = count( $pages );

		for ( $i = 0; $i < $length; $i++ ) {
			$page = & $pages[ $i ];

			if ( $this->wplegalpages_pro_ancestor_excluded( $page, $excluded_ids, $pages ) ) {
				$excluded_ids[] = $page->ID;
			}
		}

		$delete_ids = array_unique( $excluded_ids );

		for ( $i = 0; $i < $length; $i++ ) {
			$page = & $pages[ $i ];

			if ( in_array( $page->ID, $delete_ids, true ) ) {
				unset( $pages[ $i ] );
			}
		}

		if ( ! is_array( $pages ) ) {
			$pages = (array) $pages;
		}
		$pages = array_values( $pages );

		return $pages;
	}

	/**
	 * Adds metabox for to exclude pages.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_admin_init() {
		add_meta_box( 'lp_admin_meta_box', esc_attr__( 'Exclude Pages', 'wplegalpages' ), array( $this, 'wplegalpages_pro_admin_sidebar_wp25' ), 'page', 'side', 'low' );
	}

	/**
	 * Get excluded page ids.
	 *
	 * @return array
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_get_excluded_ids() {
		$exclude_ids_str = get_option( WLPP_OPTION_NAME );
		if ( empty( $exclude_ids_str ) ) {
			return array();
		}
		return explode( WLPP_OPTION_SEP, $exclude_ids_str );
	}

	/**
	 * Returns page.
	 *
	 * @param int $page_id Page ID.
	 * @param int $pages Pages.
	 * @return bool
	 */
	public function wplegalpages_pro_get_page( $page_id, $pages ) {
		$length = count( $pages );
		for ( $i = 0; $i < $length; $i++ ) {
			$page = & $pages[ $i ];
			if ( $page_id === $page->ID ) {
				return $page;
			}
		}
		return false;
	}

	/**
	 * Get ancestor excluded ancestors.
	 *
	 * @param ArrayObject $page Page.
	 * @param Array       $excluded_ids Excluded Ids.
	 * @param ArrayObject $pages Pages.
	 *
	 * @return bool|int
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_ancestor_excluded( $page, $excluded_ids, $pages ) {
		$parent = $this->wplegalpages_pro_get_page( $page->post_parent, $pages );
		if ( ! $parent ) {
			return false;
		}
		if ( in_array( $parent->ID, $excluded_ids, true ) ) {
			return (int) $parent->ID;
		}
		if ( 0 === $parent->ID ) {
			return false;
		}
		return $this->wplegalpages_pro_ancestor_excluded( $parent, $excluded_ids, $pages );
	}

	/**
	 * Get nearest excluded ancestors.
	 *
	 * @return bool|int
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_nearest_excluded_ancestor() {
		global $post_ID, $wpdb;
		if ( ! $post_ID ) {
			return false;
		}
		$excluded_ids = $this->wplegalpages_pro_get_excluded_ids();
		$sql          = "SELECT ID, post_parent FROM $wpdb->posts WHERE post_type = 'page'";
		$pages        = $wpdb->get_results( $sql ); // phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared,WordPress.DB.DirectDatabaseQuery,WordPress.DB.DirectDatabaseQuery.NoCaching
		$parent       = $this->wplegalpages_pro_get_page( $post_ID, $pages );
		return $this->wplegalpages_pro_ancestor_excluded( $parent, $excluded_ids, $pages );
	}

	/**
	 * Get included pages.
	 *
	 * @return bool
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_this_page_included() {
		global $post_ID;
		if ( ! $post_ID ) {
			return true;
		}
		$excluded_ids = $this->wplegalpages_pro_get_excluded_ids();
		if ( empty( $excluded_ids ) ) {
			return true;
		}
		return ! in_array( $post_ID, $excluded_ids, true );
	}

	/**
	 * Check if has menu.
	 *
	 * @return bool
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_has_menu() {
		if ( ! function_exists( 'wp_get_nav_menus' ) ) {
			return false;
		}
		$menus = wp_get_nav_menus();
		foreach ( $menus as $menu_maybe ) {
			if ( wp_get_nav_menu_items( $menu_maybe->term_id ) ) {
				return true;
			}
		}
	}

	/**
	 * Metabox display for exclude pages.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_admin_sidebar_wp25() {
		wp_nonce_field( 'wplegal_save_custom_metabox_inclusion', 'wplegal_custom_inclusion_meta_data' );
		$nearest_excluded_ancestor = $this->wplegalpages_pro_nearest_excluded_ancestor();
		echo '	<div id="excludepagediv" class="new-admin-wp25">';
		echo '		<div class="outer"><div class="inner">';
		echo '		<p><label for="lp_this_page_included" class="selectit">';
		echo '		<input ';
		echo '			type="checkbox" ';
		echo '			name="lp_this_page_included" ';
		echo '			id="lp_this_page_included" ';
		if ( $this->wplegalpages_pro_this_page_included() ) {
			echo 'checked="checked"';
		}
		echo ' />';
		echo '			' . esc_attr__( 'Include this page in lists of pages', 'wplegalpages' ) . '</label>';
		echo '		<input type="hidden" name="lp_ctrl_present" value="1" /></p>';
		if ( false !== $nearest_excluded_ancestor ) {
			$nearest_link       = 'post.php?action=edit&amp;post=' . $nearest_excluded_ancestor;
			$nearest_link_title = 'edit the excluded ancestor';
			echo '<p class="ep_exclude_alert"><em>';
			printf(
				/* translators: 1: post edit path 2: a tag title */
				esc_attr__( 'N.B. An ancestor of this page is excluded, so this page is too (<a href="%1$s" title="%2$s">edit ancestor</a>).', 'wplegalpages' ),
				esc_attr( $nearest_link ),
				esc_attr( $nearest_link_title )
			);
			echo '</em></p>';
		}
		if ( $this->wplegalpages_pro_has_menu() ) {
			echo '<p id="ep_custom_menu_alert"><em>';
			if ( current_user_can( 'edit_theme_options' ) ) {
				printf(
					/* translators: 1: menu php file */
					esc_attr__( 'N.B. This page can still appear in explicitly created <a href="%1$s">menus</a> (<a id="ep_toggle_more" href="#ep_explain_more">explain more</a>)', 'wplegalpages' ),
					'nav-menus.php'
				);
			} else {
				esc_attr_e( 'N.B. This page can still appear in explicitly created menus (<a id="ep_toggle_more" href="#ep_explain_more">explain more</a>)', 'wplegalpages' );
			}
			echo '</em></p>';
			echo '<div id="ep_explain_more"><p>';
			if ( current_user_can( 'edit_theme_options' ) ) {
				printf(
					/* translators: 1: menu php file */
					esc_attr__( 'WordPress provides a simple function for you to maintain your site <a href="%1$s">menus</a>. If you create a menu which includes this page, the checkbox above will not have any effect on the visibility of that menu item.', 'wplegalpages' ),
					'nav-menus.php'
				);
			} else {
				esc_attr_e( 'WordPress provides a simple function for you to maintain the site menus, which your site administrator is using. If a menu includes this page, the checkbox above will not have any effect on the visibility of that menu item.', 'wplegalpages' );
			}
			echo '</p><p>';
			echo esc_attr__( 'If you think you no longer need the Exclude Pages plugin you should talk to your WordPress administrator about disabling it.', 'wplegalpages' );
			echo '</p></div>';
		}
		echo '		</div><!-- .inner --></div><!-- .outer -->';
		echo '	</div><!-- #excludepagediv -->';
	}

	/**
	 * Sets options.
	 *
	 * @param string $name Option name.
	 * @param string $value Option value.
	 * @param string $description Description.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_set_option( $name, $value, $description ) {
		delete_option( $name );
		add_option( $name, $value );
	}

	/**
	 * Update meta data related to exclude pages.
	 *
	 * @param int $post_ID Post ID.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_update_exclusions( $post_ID ) {
		if ( isset( $_POST['wplegal_custom_inclusion_meta_data'] ) ) {
			check_admin_referer( 'wplegal_save_custom_metabox_inclusion', 'wplegal_custom_inclusion_meta_data' );
		}
		$exclude_this_page = isset( $_POST['lp_this_page_included'] ) ? ! (bool) sanitize_text_field( wp_unslash( $_POST['lp_this_page_included'] ) ) : false;
		$ctrl_present      = isset( $_POST['lp_ctrl_present'] ) ? (bool) sanitize_text_field( wp_unslash( $_POST['lp_ctrl_present'] ) ) : false;
		if ( ! $ctrl_present ) {
			return;
		}
		$excluded_ids = $this->wplegalpages_pro_get_excluded_ids();
		if ( $exclude_this_page ) {
			array_push( $excluded_ids, $post_ID );
			$excluded_ids = array_unique( $excluded_ids );
		}
		if ( ! $exclude_this_page ) {
			$index = array_search( $post_ID, $excluded_ids, true );
			if ( false !== $index ) {
				unset( $excluded_ids[ $index ] );
			}
		}
		$excluded_ids_str = implode( ',', $excluded_ids );
		$this->wplegalpages_pro_set_option( WLPP_OPTION_NAME, $excluded_ids_str, esc_attr__( 'Comma separated list of post and page IDs to exclude when returning pages from the get_pages function.', 'wplegalpages' ) );
	}

	/**
	 * Save Legal Pages Pro Admin settings.
	 *
	 * @param Array $lp_general General settings.
	 * @param Array $data Data.
	 *
	 * @return mixed
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_save_settings( $lp_general, $data ) {
		$lp_general['facebook-url']         = isset( $data['lp-facebook-url'] ) ? sanitize_text_field( esc_attr( $data['lp-facebook-url'] ) ) : '';
		$lp_general['google-url']           = isset( $data['lp-google-url'] ) ? sanitize_text_field( esc_attr( $data['lp-google-url'] ) ) : '';
		$lp_general['twitter-url']          = isset( $data['lp-twitter-url'] ) ? sanitize_text_field( esc_attr( $data['lp-twitter-url'] ) ) : '';
		$lp_general['linkedin-url']         = isset( $data['lp-linkedin-url'] ) ? sanitize_text_field( esc_attr( $data['lp-linkedin-url'] ) ) : '';
		$lp_general['leave-url']            = isset( $data['lp-leave-url'] ) ? sanitize_text_field( esc_attr( $data['lp-leave-url'] ) ) : '';
		$lp_general['date']                 = isset( $data['lp-date'] ) ? sanitize_text_field( esc_attr( $data['lp-date'] ) ) : '';
		$lp_general['recipient-party']      = isset( $data['lp-recipient-party'] ) ? sanitize_text_field( esc_attr( $data['lp-recipient-party'] ) ) : '';
		$lp_general['disclosing-party']     = isset( $data['lp-disclosing-party'] ) ? sanitize_text_field( esc_attr( $data['lp-disclosing-party'] ) ) : '';
		$lp_general['affiliate-disclosure'] = isset( $data['lp-affiliate-disclosure'] ) ? sanitize_text_field( esc_attr( $data['lp-affiliate-disclosure'] ) ) : 0;
		$lp_general['days']                 = isset( $data['lp-days'] ) ? sanitize_text_field( esc_attr( $data['lp-days'] ) ) : '';
		$lp_general['duration']             = isset( $data['lp-duration'] ) ? sanitize_text_field( esc_attr( $data['lp-duration'] ) ) : '';
		$lp_general['search']               = isset( $data['lp-search'] ) ? sanitize_text_field( esc_attr( $data['lp-search'] ) ) : 0;
		$lp_general['generate']             = isset( $data['lp-generate'] ) ? sanitize_text_field( esc_attr( $data['lp-generate'] ) ) : 0;
		$lp_general['is_adult']             = isset( $data['lp-is_adult'] ) ? sanitize_text_field( esc_attr( $data['lp-is_adult'] ) ) : 0;
		$lp_general['disable_comments']     = 1;
		return $lp_general;
	}

	/**
	 * Adds map to templates.
	 *
	 * @return array|mixed|object
	 *
	 * @since 7.0
	 */
	public function add_map() {
		$output             = '';
		$lp_general         = get_option( 'lp_general' );
		$address_street     = isset( $lp_general['street'] ) ? $lp_general['street'] : '';
		$address_city_state = isset( $lp_general['cityState'] ) ? $lp_general['cityState'] : '';
		$address_country    = isset( $lp_general['country'] ) ? $lp_general['country'] : '';
		$prep_addr          = str_replace( ' ', '+', "'.$address_street.$address_city_state.$address_country.'" );
		$url                = 'http://maps.google.com/maps/api/geocode/json?address=' . $prep_addr . '&sensor=false';
		$response           = wp_remote_get( $url );
		$response_status    = wp_remote_retrieve_response_code( $response );
		if ( 200 === $response_status ) {
			$output = wp_remote_retrieve_body( $response );
		}
		return $output;
	}

	/**
	 * Process Legal Pages shortcodes.
	 *
	 * @param string $content Content.
	 *
	 * @return mixed
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_shortcode_content( $content ) {
		$lp_general = get_option( 'lp_general' );

		$content = str_replace( '[date]', isset( $lp_general['date'] ) ? $lp_general['date'] : '', stripslashes( $content ) );
		$content = str_replace( '[recipient-party]', isset( $lp_general['recipient-party'] ) ? $lp_general['recipient-party'] : '', stripslashes( $content ) );
		$content = str_replace( '[disclosing-party]', isset( $lp_general['disclosing-party'] ) ? $lp_general['disclosing-party'] : '', stripslashes( $content ) );
		$content = str_replace( '[days]', isset( $lp_general['days'] ) ? $lp_general['days'] : '', stripslashes( $content ) );
		$content = str_replace( '[duration]', isset( $lp_general['duration'] ) ? $lp_general['duration'] : '', stripslashes( $content ) );

		$latitude  = '';
		$longitude = '';
		$facebook  = '';
		$google    = '';
		$twitter   = '';
		$linkedin  = '';

		$pos = strpos( $content, '[latitude]' );
		if ( false !== $pos ) {
			$output = $this->add_map();

			if ( ! empty( $output->results ) ) {
				$latitude  = $output->results[0]->geometry->location->lat;
				$longitude = $output->results[0]->geometry->location->lng;
			}
			$lp_map_find   = array( '[latitude]', '[longitude]' );
			$map_dimension = array(
				'latitude'  => $latitude,
				'longitude' => $longitude,
			);
			$content       = str_replace( $lp_map_find, $map_dimension, stripslashes( $content ) );
		}

		$facebook_url = isset( $lp_general['facebook-url'] ) ? $lp_general['facebook-url'] : '';

		if ( $facebook_url ) {

			$facebook = array( 'facebook' => '<a style="width: 30px; height: 30px; display:inline-block" href="' . $facebook_url . '"><img src="' . WPLEGALPAGES_PRO_PLUGIN_URL . 'admin/images/facebook.png"/></a>' );
		}
		$lp_find_facebook = array( '[facebook]' );
		$content          = str_replace( $lp_find_facebook, $facebook, stripslashes( $content ) );

		$google_url = isset( $lp_general['google-url'] ) ? $lp_general['google-url'] : '';

		if ( $google_url ) {

			$google = array( 'google' => '<a style="width: 30px; height: 30px; display:inline-block" href="' . $google_url . '"><img src="' . WPLEGALPAGES_PRO_PLUGIN_URL . 'admin/images/google.png"/></a>' );
		}
		$lp_find_google = array( '[google]' );
		$content        = str_replace( $lp_find_google, $google, stripslashes( $content ) );

		$twitter_url = isset( $lp_general['twitter-url'] ) ? $lp_general['twitter-url'] : '';
		if ( $twitter_url ) {

			$twitter = array( 'twitter' => '<a style="width: 30px; height: 30px; display:inline-block" href="' . $twitter_url . '"><img src="' . WPLEGALPAGES_PRO_PLUGIN_URL . 'admin/images/twitter.png"/></a>' );
		}
		$lp_find_twitter = array( '[twitter]' );
		$content         = str_replace( $lp_find_twitter, $twitter, stripslashes( $content ) );

		$linkedin_url = isset( $lp_general['linkedin-url'] ) ? $lp_general['linkedin-url'] : '';
		if ( $linkedin_url ) {
			$linkedin = array( 'linkedin' => '<a style="width: 30px; height: 30px; display:inline-block" href="' . $linkedin_url . '"><img src="' . WPLEGALPAGES_PRO_PLUGIN_URL . 'admin/images/linkedin.png"/></a>' );
		}
		$lp_find_linkedin = array( '[linkedin]' );
		$content          = str_replace( $lp_find_linkedin, $linkedin, stripslashes( $content ) );
		return $content;
	}
}
