<?php
/**
 * The file that defines the core plugin class
 *
 * A class definition that includes attributes and functions used across both the
 * public-facing side of the site and the admin area.
 *
 * @link       https://club.wpeka.com
 * @since      7.0
 *
 * @package    Wplegalpages_Pro
 * @subpackage Wplegalpages_Pro/includes
 */

/**
 * The core plugin class.
 *
 * This is used to define internationalization, admin-specific hooks, and
 * public-facing site hooks.
 *
 * Also maintains the unique identifier of this plugin as well as the current
 * version of the plugin.
 *
 * @since      7.0
 * @package    Wplegalpages_Pro
 * @subpackage Wplegalpages_Pro/includes
 * @author     WPEkaClub <support@wpeka.com>
 */
class Wplegalpages_Pro {

	/**
	 * The loader that's responsible for maintaining and registering all hooks that power
	 * the plugin.
	 *
	 * @since    7.0
	 * @access   protected
	 * @var      Wplegalpages_Pro_Loader    $loader    Maintains and registers all hooks for the plugin.
	 */
	protected $loader;

	/**
	 * The unique identifier of this plugin.
	 *
	 * @since    7.0
	 * @access   protected
	 * @var      string    $plugin_name    The string used to uniquely identify this plugin.
	 */
	protected $plugin_name;

	/**
	 * The current version of the plugin.
	 *
	 * @since    7.0
	 * @access   protected
	 * @var      string    $version    The current version of the plugin.
	 */
	protected $version;

	/**
	 * Define the core functionality of the plugin.
	 *
	 * Set the plugin name and the plugin version that can be used throughout the plugin.
	 * Load the dependencies, define the locale, and set the hooks for the admin area and
	 * the public-facing side of the site.
	 *
	 * @since    7.0
	 */
	public function __construct() {
		global $table_prefix;
		if ( defined( 'WPLEGALPAGES_PRO_VERSION' ) ) {
			$this->version = WPLEGALPAGES_PRO_VERSION;
		} else {
			$this->version = '7.6';
		}
		$this->plugin_name = 'wplegalpages-pro';

		$this->plugin_url = plugin_dir_url( dirname( __FILE__ ) );

		$this->tablename  = $table_prefix . 'legal_pages';
		$this->popuptable = $table_prefix . 'lp_popups';

		$this->define_constants();
		$this->load_dependencies();
		$this->set_locale();
		$this->define_admin_hooks();
		$this->define_public_hooks();

	}

	/**
	 * Define the required constantns for this plugin.
	 *
	 * @since 7.0
	 * @access   private
	 */
	private function define_constants() {
		if ( ! defined( 'WLPP_OPTION_NAME' ) ) {
			define( 'WLPP_OPTION_NAME', 'lp_exclude_pages' );
		}
		if ( ! defined( 'WLPP_OPTION_SEP' ) ) {
			define( 'WLPP_OPTION_SEP', ',' );
		}
	}

	/**
	 * Load the required dependencies for this plugin.
	 *
	 * Include the following files that make up the plugin:
	 *
	 * - Wplegalpages_Pro_Loader. Orchestrates the hooks of the plugin.
	 * - Wplegalpages_Pro_I18n. Defines internationalization functionality.
	 * - Wplegalpages_Pro_Admin. Defines all hooks for the admin area.
	 * - Wplegalpages_Pro_Public. Defines all hooks for the public side of the site.
	 *
	 * Create an instance of the loader which will be used to register the hooks
	 * with WordPress.
	 *
	 * @since    7.0
	 * @access   private
	 */
	private function load_dependencies() {

		/**
		 * The class responsible for orchestrating the actions and filters of the
		 * core plugin.
		 */
		require_once plugin_dir_path( dirname( __FILE__ ) ) . 'includes/class-wplegalpages-pro-loader.php';

		/**
		 * The class responsible for defining internationalization functionality
		 * of the plugin.
		 */
		require_once plugin_dir_path( dirname( __FILE__ ) ) . 'includes/class-wplegalpages-pro-i18n.php';

		/**
		 * The class responsible for defining all actions that occur in the admin area.
		 */
		require_once plugin_dir_path( dirname( __FILE__ ) ) . 'admin/class-wplegalpages-pro-admin.php';

		/**
		 * The class responsible for defining all actions that occur in the public-facing
		 * side of the site.
		 */
		require_once plugin_dir_path( dirname( __FILE__ ) ) . 'public/class-wplegalpages-pro-public.php';

		$this->loader = new Wplegalpages_Pro_Loader();

	}

	/**
	 * Define the locale for this plugin for internationalization.
	 *
	 * Uses the Wplegalpages_Pro_I18n class in order to set the domain and to register the hook
	 * with WordPress.
	 *
	 * @since    7.0
	 * @access   private
	 */
	private function set_locale() {

		$plugin_i18n = new Wplegalpages_Pro_I18n();

		$this->loader->add_action( 'plugins_loaded', $plugin_i18n, 'load_plugin_textdomain' );

	}

	/**
	 * Register all of the hooks related to the admin area functionality
	 * of the plugin.
	 *
	 * @since    7.0
	 * @access   private
	 */
	private function define_admin_hooks() {

		$plugin_admin = new Wplegalpages_Pro_Admin( $this->get_plugin_name(), $this->get_version() );

		$this->loader->add_action( 'admin_enqueue_scripts', $plugin_admin, 'enqueue_styles' );
		$this->loader->add_action( 'admin_enqueue_scripts', $plugin_admin, 'enqueue_scripts' );
		$this->loader->add_action( 'wplegalpages_admin_menu', $plugin_admin, 'wplegalpages_pro_admin_menu' );
		$this->loader->add_action( 'init', $plugin_admin, 'wplegalpages_pro_init' );
		$this->loader->add_action( 'admin_init', $plugin_admin, 'wplegalpages_pro_admin_init' );
		$this->loader->add_action( 'save_post', $plugin_admin, 'wplegalpages_pro_update_exclusions' );
		$this->loader->add_action( 'wp_dashboard_setup', $plugin_admin, 'wplegalpages_pro_dashboard_setup' );
		$this->loader->add_action( 'wplegalpages_admin_settings', $plugin_admin, 'wplegalpages_pro_admin_settings', 10, 1 );
		$this->loader->add_action( 'wplegalpages_save_settings', $plugin_admin, 'wplegalpages_pro_save_settings', 10, 2 );
		$this->loader->add_filter( 'wplegalpages_shortcode_content', $plugin_admin, 'wplegalpages_pro_shortcode_content', 10, 1 );
		$this->loader->add_filter( 'wplegalpages_pages_limit', $plugin_admin, 'wplegalpages_pro_pages_limit', 10, 1 );
	}

	/**
	 * Register all of the hooks related to the public-facing functionality
	 * of the plugin.
	 *
	 * @since    7.0
	 * @access   private
	 */
	private function define_public_hooks() {

		$plugin_public = new Wplegalpages_Pro_Public( $this->get_plugin_name(), $this->get_version() );

		$this->loader->add_action( 'wp_enqueue_scripts', $plugin_public, 'enqueue_styles' );
		$this->loader->add_action( 'wp_enqueue_scripts', $plugin_public, 'enqueue_scripts' );
		$this->loader->add_action( 'wp_head', $plugin_public, 'wplegalpages_pro_disable_comments' );

	}

	/**
	 * Run the loader to execute all of the hooks with WordPress.
	 *
	 * @since    7.0
	 */
	public function run() {
		$this->loader->run();
	}

	/**
	 * The name of the plugin used to uniquely identify it within the context of
	 * WordPress and to define internationalization functionality.
	 *
	 * @since     7.0
	 * @return    string    The name of the plugin.
	 */
	public function get_plugin_name() {
		return $this->plugin_name;
	}

	/**
	 * The reference to the class that orchestrates the hooks with the plugin.
	 *
	 * @since     7.0
	 * @return    Wplegalpages_Pro_Loader    Orchestrates the hooks of the plugin.
	 */
	public function get_loader() {
		return $this->loader;
	}

	/**
	 * Retrieve the version number of the plugin.
	 *
	 * @since     7.0
	 * @return    string    The version number of the plugin.
	 */
	public function get_version() {
		return $this->version;
	}

	/**
	 * Enqueues editor specific javascript files.
	 *
	 * @since 7.0
	 */
	public function wplegalpages_pro_enqueue_editor() {
		wp_enqueue_script( 'common' );
		wp_enqueue_script( 'jquery-affect' );
		wp_admin_css( 'thickbox' );
		wp_print_scripts( 'post' );
		wp_print_scripts( 'media-upload' );
		wp_print_scripts( 'jquery' );
		wp_print_scripts( 'jquery-ui-core' );
		wp_print_scripts( 'jquery-ui-tabs' );
		wp_print_scripts( 'editor' );
		wp_enqueue_style( 'bootstrap-min', $this->plugin_url . 'admin/css/bootstrap.min.css', array(), $this->version );
		wp_enqueue_style( 'style', $this->plugin_url . 'admin/css/style' . WPLPP_SUFFIX . '.css', array(), $this->version );
		wp_enqueue_script( 'jscolor', $this->plugin_url . 'admin/js/jscolor/jscolor' . WPLPP_SUFFIX . '.js', array(), $this->version, true );
		wp_enqueue_script( 'jquery-min', $this->plugin_url . 'admin/js/jquery.min.js', array(), $this->version, true );
		wp_enqueue_script( 'bootstrap-min-js', $this->plugin_url . 'admin/js/bootstrap.min.js', array(), $this->version, true );
		wp_enqueue_script( 'tooltip-js', $this->plugin_url . 'admin/js/tooltip' . WPLPP_SUFFIX . '.js', array(), $this->version, true );
		wp_print_scripts( 'wplink' );
		wp_print_styles( 'wplink' );
		add_thickbox();
		wp_admin_css();
		wp_enqueue_script( 'utils' );
		do_action( 'admin_print_styles_post_php' );
		do_action( 'admin_print_styles' );
	}

}
